/* File to HDF                                                              */

/*--------------------------------------------------------------------------*/
/* system includes                                                          */
/*--------------------------------------------------------------------------*/

	#include <stdlib.h>
	#include <string.h>
	#include <stdio.h>

/*--------------------------------------------------------------------------*/
/* macros and constants                                                     */
/*--------------------------------------------------------------------------*/

	/* read big endian longword */

	#define MAKE_ID(p) (((p)[0] << 24) | ((p)[1] << 16) | ((p)[2] << 8) | ((p)[3]))
	#define GET_ID(p,o) (((p)[o] << 24) | ((p)[(o)+1] << 16) | ((p)[(o)+2] << 8) | ((p)[(o)+3]))


	/* RigidDiskBlock */

	#define	IDNAME_RIGIDDISK	0x5244534B	/* 'RDSK' */

	#define rdb_ID(p)                 MAKE_ID(p)
	#define rdb_SummedLongs(p)        GET_ID(p,4)
	#define rdb_ChkSum(p)             GET_ID(p,8)
	#define rdb_HostID(p)             GET_ID(p,12)
	#define rdb_BlockBytes(p)         GET_ID(p,16)
	#define rdb_Flags(p)              GET_ID(p,20)
	#define rdb_BadBlockList(p)       GET_ID(p,24)
	#define rdb_PartitionList(p)      GET_ID(p,28)
	#define rdb_FileSysHeaderList(p)  GET_ID(p,32)
	#define rdb_DriveInit(p)          GET_ID(p,36)
	/*      rdb_Reserved1[6] */
	#define rdb_Cylinders(p)          GET_ID(p,64)
	#define rdb_Sectors(p)            GET_ID(p,68)
	#define rdb_Heads(p)              GET_ID(p,72)
	#define rdb_Interleave(p)         GET_ID(p,76)
	#define rdb_Park(p)               GET_ID(p,80)
	/*      rdb_Reserved2[3] */
	#define rdb_WritePreComp(p)       GET_ID(p,96)
	#define rdb_ReducedWrite(p)       GET_ID(p,100)
	#define rdb_StepRate(p)           GET_ID(p,104)
	/*      rdb_Reserved3 */
	#define rdb_RDBBlocksLo(p)        GET_ID(p,128)
	#define rdb_RDBBlocksHi(p)        GET_ID(p,132)
	#define rdb_LoCylinder(p)         GET_ID(p,136)
	#define rdb_HiCylinder(p)         GET_ID(p,140)
	#define rdb_CylBlocks(p)          GET_ID(p,144)
	#define rdb_AutoParkSeconds(p)    GET_ID(p,148)
	#define rdb_HighRDSKBlock(p)      GET_ID(p,152)
	/*      rdb_Reserved4 */
	#define rdb_DiskVendor(p)         (&(p)[160])
	#define rdb_DiskProduct(p)        (&(p)[168])
	#define rdb_DiskRevision(p)       (&(p)[184])
	#define rdb_ControllerVendor(p)   (&(p)[188])
	#define rdb_ControllerProduct(p)  (&(p)[196])
	#define rdb_ControllerRevision(p) (&(p)[212])
	#define rdb_DriveInitName(p)      (&(p)[216])

	/* PartitionBlock */

	#define	IDNAME_PARTITION	0x50415254	/* 'PART' */

	#define pb_ID(p)              MAKE_ID(p)
	#define pb_SummedLongs(p)     GET_ID(p,4)
	#define pb_ChkSum(p)          GET_ID(p,8)
	#define pb_HostID(p)          GET_ID(p,12)
	#define pb_Next(p)            GET_ID(p,16)
	#define pb_Flags(p)           GET_ID(p,20)
	/*      pb_Reserved1[2] */
	#define pb_DevFlags(p)        GET_ID(p,32)
	#define pb_DriveName(p)       (&(p)[36])
	/*      pb_Reserved2[15] */
	#define pb_Environment(p)     (&(p)[128])
	/*      pb_EReserved[12] */

	/* DosEnvec */

	#define de_TableSize(p)       MAKE_ID(p)
	#define de_SizeBlock(p)       GET_ID(p,4)
	#define de_SecOrg(p)          GET_ID(p,8)
	#define de_Surfaces(p)        GET_ID(p,12)
	#define de_SectorPerBlock(p)  GET_ID(p,16)
	#define de_BlocksPerTrack(p)  GET_ID(p,20)
	#define de_Reserved(p)        GET_ID(p,24)
	#define de_PreAlloc(p)        GET_ID(p,28)
	#define de_Interleave(p)      GET_ID(p,32)
	#define de_LowCyl(p)          GET_ID(p,36)
	#define de_HighCyl(p)         GET_ID(p,40)
	#define de_NumBuffers(p)      GET_ID(p,44)
	#define de_BufMemType(p)      GET_ID(p,48)
	#define de_MaxTransfer(p)     GET_ID(p,52)
	#define de_Mask(p)            GET_ID(p,56)
	#define de_BootPri(p)         GET_ID(p,60)
	#define de_DosType(p)         GET_ID(p,64)
	#define de_Baud(p)            GET_ID(p,68)
	#define de_Control(p)         GET_ID(p,72)
	#define de_BootBlocks(p)      GET_ID(p,76)

/*--------------------------------------------------------------------------*/
/* new types and structures                                                 */
/*--------------------------------------------------------------------------*/

	struct args {
		char *from;
		char *to;
		long part;
		long blksz;
		long buffers;
		long list;
		long verbose;
		long ml;
		char *device;
		long unit;
		};

	struct part {
		char name[32];
		unsigned long TableSize;
		unsigned long SizeBlock;
		unsigned long SecOrg;
		unsigned long Surfaces;
		unsigned long SectorPerBlock;
		unsigned long BlocksPerTrack;
		unsigned long Reserved;
		unsigned long PreAlloc;
		unsigned long Interleave;
		unsigned long LowCyl;
		unsigned long HighCyl;
		unsigned long NumBuffers;
		unsigned long BufMemType;
		unsigned long MaxTransfer;
		unsigned long Mask;
		         long BootPri;
		unsigned long DosType;
		unsigned long Baud;
		unsigned long Control;
		unsigned long BootBlocks;
		};


/*--------------------------------------------------------------------------*/
/* global variables                                                         */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* prototypes                                                               */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* search a RDB in a file                                                   */
/*--------------------------------------------------------------------------*/

	short search_rdb (FILE *in,unsigned char *buffer,long blksz)

	{
	short found = 0;
	long num;

	fseek (in,0,SEEK_SET);

	for (num = 0; num < 16 && !found; num++)
		{
		fread (buffer,blksz,1,in);
		if (rdb_ID(buffer) == IDNAME_RIGIDDISK)
			found = 1;
		}

	return (found);
	}

/*--------------------------------------------------------------------------*/
/* copy BCPL string into C string                                           */
/*--------------------------------------------------------------------------*/

	void bstrcpy (unsigned char *dest,unsigned char *src)

	{
	int len;

	len = *src++;
	strncpy (dest,src,len);
	dest[len] = '\0';
	}

/*--------------------------------------------------------------------------*/
/* read partition block from file                                           */
/*--------------------------------------------------------------------------*/

	void read_part (FILE *in,unsigned char *buffer,long blksz,long block,struct part *part)

	{
	unsigned char *env;

	fseek (in,block * blksz,SEEK_SET);
	fread (buffer,blksz,1,in);

	bstrcpy (part->name,pb_DriveName(buffer));

	env = pb_Environment(buffer);

	part->TableSize      = de_TableSize(env);
	part->SizeBlock      = de_SizeBlock(env);
	part->SecOrg         = de_SecOrg(env);
	part->Surfaces       = de_Surfaces(env);
	part->SectorPerBlock = de_SectorPerBlock(env);
	part->BlocksPerTrack = de_BlocksPerTrack(env);
	part->Reserved       = de_Reserved(env);
	part->PreAlloc       = de_PreAlloc(env);
	part->Interleave     = de_Interleave(env);
	part->LowCyl         = de_LowCyl(env);
	part->HighCyl        = de_HighCyl(env);
	part->NumBuffers     = de_NumBuffers(env);
	part->BufMemType     = de_BufMemType(env);
	part->MaxTransfer    = de_MaxTransfer(env);
	part->Mask           = de_Mask(env);
	part->BootPri        = de_BootPri(env);
	part->DosType        = de_DosType(env);
	part->Baud           = de_Baud(env);
	part->Control        = de_Control(env);
	part->BootBlocks     = de_BootBlocks(env);
	}

/*--------------------------------------------------------------------------*/
/* verify command arguments                                                 */
/*--------------------------------------------------------------------------*/

	short read_args (int argc,char **argv,struct args *args)

	{
	int i;
	short err = 0;

	memset (args,0,sizeof(struct args));       /* clear arguments */

	for (i = 1; !err && i < argc; i++)         /* check syntax and fill arguments */
		{
		if (argv[i][0] == '-')
			{
			switch (argv[i][1])
				{
			case 'f':                         /* input file */
				args->from = argv[++i];
				break;
			case 'o':                         /* output file */
				args->to = argv[++i];
				break;
			case 'p':                         /* partition number */
				args->part = atoi(argv[++i]);
				break;
			case 's':                         /* block size */
				args->blksz = atoi(argv[++i]);
				break;
			case 'l':                         /* list */
				args->list = 1;
				if (argv[i][2] == 'v')        /* list verbose */
					args->verbose = 1;
				break;
			case 'm':                         /* mount list */
				args->ml = 1;
				break;
			case 'd':                         /* device name */
				args->device = argv[++i];
				break;
			case 'u':                         /* unit number */
				args->unit = atoi(argv[++i]);
				break;
			case 'h':                         /* show help */
				err = 1;
				break;
			default:
				printf ("invalid option: %s\n",argv[i]);
				err = 1;
				}
			}
		else
			{
			printf ("not an option: %s\n",argv[i]);
			err = 1;
			}
		}

	if (err == 0)                              /* check logic */
		{
		if (!args->from)
			{
			printf ("input file is missing\n");
			err = 1;
			}
		else if (!args->list && !args->to)
			{
			printf ("output file is missing\n");
			err = 1;
			}
		}

	if (err)                                   /* show usage */
		{
		printf ("\nUsage: FileToHDF <options>\n"
		        "\nOptions:\n\n"
		        "  -f name     (required) name of input file, usualy something like /dev/hda\n"
		        "  -o name     (required if -l is not specified) name of output file.\n"
		        "  -p number   number of the partition to be dumped (default: 1 or all for -m)\n"
		        "  -s size     block size in bytes (default: 512)\n"
		        "  -b size     buffer size in blocks (default: 1000)\n"
		        "  -l          only list partitions (don't write output file)\n"
		        "  -lv         list more info\n"
		        "  -m          write mount list instead of HDF\n"
		        "  -d device   name of the device for mountlist (default: scsi.device)\n"
		        "  -u unit     unit number for mountlist (default: 0)\n"
		        "  -h          show this help\n"
		        "\nOptions must be followed by a space character (-f/dev/hda is not allowed)\n"
		        );
		}

	return (err == 0);
	}


/*--------------------------------------------------------------------------*/
/* write mount list entry for one partition                                 */
/*--------------------------------------------------------------------------*/


	void write_ml (FILE *out,struct part *part,char *device,long unit)

	{
	fprintf (out,"%s:\n",part->name);
	fprintf (out,"  Device          = \"%s\"\n",device ? device : "scsi.device");
	fprintf (out,"  Unit            = %ld\n",unit);
	fprintf (out,"  SectorSize      = %lu\n",part->SizeBlock * 4);
	fprintf (out,"  Surfaces        = %lu\n",part->Surfaces);
	fprintf (out,"  SectorsPerBlock = %lu\n",part->SectorPerBlock);
	fprintf (out,"  BlocksPerTrack  = %lu\n",part->BlocksPerTrack);
	fprintf (out,"  Reserved        = %lu\n",part->Reserved);
	fprintf (out,"  Interleave      = %lu\n",part->Interleave);
	fprintf (out,"  LowCyl          = %lu\n",part->LowCyl);
	fprintf (out,"  HighCyl         = %lu\n",part->HighCyl);
	fprintf (out,"  Buffers         = %lu\n",part->NumBuffers);
	fprintf (out,"  BufMemType      = %lu\n",part->BufMemType);
	fprintf (out,"  MaxTransfer     = 0x%lx\n",part->MaxTransfer);
	fprintf (out,"  Mask            = 0x%08lx\n",part->Mask);
	fprintf (out,"  BootPri         = %ld\n",part->BootPri);
	fprintf (out,"  DosType         = 0x%08lx\n",part->DosType);
	fprintf (out,"#\n");
	}


/*--------------------------------------------------------------------------*/
/* copy one partition                                                       */
/*--------------------------------------------------------------------------*/

	void copy_part (FILE *in,FILE *out,long first,long last,long blksz,char *buffer,long nbuff)

	{
	long rest;

	fseek (in,first * blksz,SEEK_SET);

	rest = (last + 1 - first);

	while (rest > nbuff)
		{
		printf ("\rblocks to go: %8d",rest); 
		fflush(stdout);
		fread (buffer,blksz,nbuff,in);
		fwrite (buffer,blksz,nbuff,out);
		rest -= nbuff;
		}

	if (rest > 0)
		{
		fread (buffer,blksz,rest,in);
		fwrite (buffer,blksz,rest,out);
		}

	printf ("\rblocks to go:        0\n"); 
	}

/*--------------------------------------------------------------------------*/
/* main program                                                             */
/*--------------------------------------------------------------------------*/

	int main (int argc,char *argv[])

	{
	struct args args;
	FILE *in;
	FILE *out;
	long blksz;
	long nbuff;
	unsigned char *buffer;
	unsigned long block;
	struct part part;
	long partnum;
	long i;
	long blkpercyl;

	if (!read_args (argc,argv,&args))
		return (20);

	in = fopen (args.from,"rb");
	if (!in)
		{
		printf ("cannot open input file\n");
		return (20);
		}

	blksz = (args.blksz ? args.blksz : 512);
	nbuff = (args.buffers ? args.buffers : 1000);

	buffer = malloc (blksz * nbuff);
	if (!buffer)
		{
		fclose (in);
		printf ("not enough memory or invalid block size\n");
		return (20);
		}

	if (!search_rdb (in,buffer,blksz))
		{
		free (buffer);
		fclose (in);
		printf ("file does not contain an Amiga RDB\n");
		return (20);
		}

	blkpercyl = rdb_Sectors(buffer) * rdb_Heads(buffer);

	if (args.list)
		{
		if (args.verbose)
			{
			printf ("BlockBytes         = %ld\n",rdb_BlockBytes(buffer));
			printf ("Flags              = %08lx\n",rdb_Flags(buffer));
			printf ("Cylinders          = %ld\n",rdb_Cylinders(buffer));
			printf ("Sectors            = %ld\n",rdb_Sectors(buffer));
			printf ("Heads              = %ld\n",rdb_Heads(buffer));
			printf ("Interleave         = %ld\n",rdb_Interleave(buffer));
			printf ("Park               = %ld\n",rdb_Park(buffer));
			printf ("CylBlocks          = %ld\n",rdb_CylBlocks(buffer));
			printf ("DiskVendor         = %.8s\n",rdb_DiskVendor(buffer));
			printf ("DiskProduct        = %.16s\n",rdb_DiskProduct(buffer));
			printf ("DiskRevision       = %.4s\n",rdb_DiskRevision(buffer));
			printf ("ControllerVendor   = %.8s\n",rdb_ControllerVendor(buffer));
			printf ("ControllerProduct  = %.16s\n",rdb_ControllerProduct(buffer));
			printf ("ControllerRevision = %.4s\n",rdb_ControllerRevision(buffer));
			}
		else
			{
			printf ("disk info: %.8s%.16s%.4s\n",rdb_DiskVendor(buffer),rdb_DiskProduct(buffer),rdb_DiskRevision(buffer));
			printf ("number of cylinders : %lu\n",rdb_Cylinders(buffer));
			printf ("tracks per cylinder : %lu\n",rdb_Heads(buffer));
			printf ("sectors per track   : %lu\n",rdb_Sectors(buffer));
			}

		for (i = 1, block = rdb_PartitionList(buffer); block != 0xffffffff; block = pb_Next(buffer), i++)
			{
			read_part (in,buffer,blksz,block,&part);
			if (args.verbose)
				{
				printf ("\nPartition %s:\n",part.name);
				printf ("  TableSize      = %lu\n",part.TableSize);
				printf ("  SizeBlock      = %lu\n",part.SizeBlock * 4);
				printf ("  SecOrg         = %lu\n",part.SecOrg);
				printf ("  Surfaces       = %lu\n",part.Surfaces);
				printf ("  SectorPerBlock = %lu\n",part.SectorPerBlock);
				printf ("  BlocksPerTrack = %lu\n",part.BlocksPerTrack);
				printf ("  Reserved       = %lu\n",part.Reserved);
				printf ("  PreAlloc       = %lu\n",part.PreAlloc);
				printf ("  Interleave     = %lu\n",part.Interleave);
				printf ("  LowCyl         = %lu\n",part.LowCyl);
				printf ("  HighCyl        = %lu\n",part.HighCyl);
				printf ("  NumBuffers     = %lu\n",part.NumBuffers);
				printf ("  BufMemType     = %lu\n",part.BufMemType);
				printf ("  MaxTransfer    = 0x%lx\n",part.MaxTransfer);
				printf ("  Mask           = 0x%08lx\n",part.Mask);
				printf ("  BootPri        = %ld\n",part.BootPri);
				printf ("  DosType        = 0x%08lx\n",part.DosType);
				printf ("  Baud           = %lu\n",part.Baud);
				printf ("  Control        = %lu\n",part.Control);
				printf ("  BootBlocks     = %lu\n",part.BootBlocks);
				}
			else
				{
				printf ("<%3lu> %-6s  start cyl: %6lu   size: %10lu bytes   dos type: 0x%08x\n",
					i,part.name,part.LowCyl,
					part.BlocksPerTrack * part.Surfaces * (part.HighCyl + 1 - part.LowCyl) * part.SizeBlock * 4,
					part.DosType);
				}
			}
		}
	else
		{
		out = fopen (args.to,(args.ml ? "w" : "wb"));
		if (!out)
			{
			free (buffer);
			fclose (in);
			printf ("cannot open output file\n");
			return (20);
			}

		if (args.ml)
			{
			if (!args.part)
				{
				for (block = rdb_PartitionList(buffer); block != 0xffffffff; block = pb_Next(buffer))
					{
					read_part (in,buffer,blksz,block,&part);
					write_ml (out,&part,args.device,args.unit);
					}
				}
			else
				{
				for (i = 0, block = rdb_PartitionList(buffer); i < args.part && block != 0xffffffff; block = pb_Next(buffer), i++)
					{
					read_part (in,buffer,blksz,block,&part);
					}

				if (i == args.part && block != 0xffffffff)
					{
					write_ml (out,&part,args.device,args.unit);
					}
				else
					printf ("partition %ld does not exist\n",args.part);
				}
			}
		else
			{
			partnum = (args.part ? args.part : 1);

			i = 0;
			for (block = rdb_PartitionList(buffer); block != 0xffffffff; block = pb_Next(buffer))
				{
				read_part (in,buffer,blksz,block,&part);
				i ++;
				if (i >= partnum)
					break;
				}

			if (i == partnum && block != 0xffffffff)
				{
				copy_part (in,out,blkpercyl * part.LowCyl,blkpercyl * (part.HighCyl + 1) - 1,blksz,buffer,nbuff);
				}
			else
				printf ("partition %ld does not exist\n",partnum);
			}

		fclose (out);
		}

	free (buffer);
	fclose (in);
	return (0);
	}

/*--------------------------------------------------------------------------*/
/* end of the source text                                                   */
/*--------------------------------------------------------------------------*/
